/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: geocat.c,v 1.3 1996/07/02 16:29:38 deaven Exp $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "dmd/geometry.h"
#include "dmd/neighbor.h"
#include "dmd/memory.h"

char *usage = "usage: %s [options] [geometry [geometry...]]\n"
  "\t-o output-file              -t x,y,z (translate)\n"
  "\t-f or-flag                  -F and-flag\n"
  "\t-c (center)                 -O (standard orientation)\n"
  "\t-s (sort by box coord.)     -S (sort by physical coord.)\n"
  "\t-b (output box format)      -x (output XYZ format)\n"
  "\t-M (compute moments of inertia)\n"
  "\t-n rc  (print neighbor list)\n"
  ;

enum {GENERAL, BOX, XYZ};
#define SMALL 1.e-10

static void translate(char *s, Geometry *g)
{
  int i;
  double dx[3], x[3];
  if(sscanf(s, "%lf,%lf,%lf", dx, dx+1, dx+2)!=3) {
    fprintf(stderr, "cannot translate by %s\n", s);
    exit(1);
  }
  for(i=0; i<(g->n); ++i) {
    (void) mmul3(g->box, g->x+3*i, x);
    x[0] += dx[0];
    x[1] += dx[1];
    x[2] += dx[2];
    (void) mmul3(g->boxi, x, g->x+3*i);
  }
}

static Geometry *gcmp;
static int sort_phys = 0;
static int cmp(const void *a, const void *b)
{
	int ia = *((int *) a), ib = *((int *) b);
	double *xb = gcmp->x + ia * 3;
	double *yb = gcmp->x + ib * 3;
	double *x, *y, xp[3], yp[3];

	if(gcmp->flags != NULL)
	{
	  if(gcmp->flags[ia] > gcmp->flags[ib]) return 1;
	  if(gcmp->flags[ia] < gcmp->flags[ib]) return -1;
	}

	if(sort_phys) {
	  x = mmul3(gcmp->box, xb, xp);
	  y = mmul3(gcmp->box, yb, yp);
	} else {
	  x = xb;
	  y = yb;
	}

	if(fabs(x[2] - y[2]) > SMALL) {
	  return (x[2] > y[2] ? 1 : -1);
	} else if(fabs(x[1] - y[1]) > SMALL) {
	  return (x[1] > y[1] ? 1 : -1);
	} else if(fabs(x[0] - y[0]) > SMALL) {
	  return (x[0] > y[0] ? 1 : -1);
	} else {
	  return 0;
	}
}

/* Call Gcat() and other things to manipulate a geometry
 * specification
 */
void main(int argc, char **argv)
{
  Geometry *g = NULL, *dg = NULL, *new = NULL;
  int i, j, arg;
  int oformat = GENERAL, sort = 0, center = 0;
  int flag = -1, Flag = -1, Orient = 0;
  int Mflag = 0;
  double nflag = -1.0;
  char *fn = NULL, *ofn = "-";
  FILE *f = NULL;

  for(arg=1; arg<argc; ++arg) {
    if(argv[arg][0] == '-') {
      switch(argv[arg][1]) {
	case 'o':
	  ofn = argv[++arg];
	  break;
	case 'O':
	  Orient = 1;
	  break;
	case 'f':
	  sscanf(argv[++arg], "%d", &flag);
	  break;
	case 'F':
	  sscanf(argv[++arg], "%d", &Flag);
	  break;
	case 'n':
	  sscanf(argv[++arg], "%lf", &nflag);
	  break;
	case 'b':
	  oformat = BOX;
	  break;
	case 'M':
	  Mflag = 1;
	  break;
	case 'x':
	  oformat = XYZ;
	  break;
	case 't':
	  translate(argv[++arg], g); break;
	case 's':
	  sort = 1; sort_phys = 0; break;
	case 'S':
	  sort = 1; sort_phys = 1; break;
	case 'c':
	  center = 1; break;
	default:
	  fprintf(stderr, usage, argv[0]);
	  exit(1);
      }
    } else { /* geometry filename */
      fn = argv[arg];
      f = (strcmp(fn, "-") ? fopen(fn, "r") : stdin);
      if(f == NULL) {
	fprintf(stderr, "%s: cannot open %s\n", argv[0], fn);
	exit(1);
      }
      dg = Gread(f);
      if(f != stdin) fclose(f);

      if(dg == NULL) {
	fprintf(stderr, "error reading %s\n", fn);
	exit(1);
      }

      if(g == NULL) {
	g = dg;
      } else {
	new = Gcat(NULL, g, dg);
	Gfree(g);
	Gfree(dg);
	g = new;
      }
    }
  }

  if (g == NULL) { /* no geometry -- try to read stdin */
    g = Gread(stdin);
  }

  if(g == NULL) {
    fprintf(stderr, "%s: need geometry\n", argv[0]);
    exit(1);
  }

  if(center) GboxCenter(g);

  if(sort) {
  	int *base = imalloc(g->n);
  	for(i=0; i<(g->n); ++i) base[i] = i;
  	gcmp = Gcpy(NULL, g);
  	qsort(base, (size_t) g->n, sizeof(int), cmp);
  	for(i=0; i<(g->n); ++i) {
	  myfree(g->aname[i]);
	  g->aname[i] = gcmp->aname[base[i]];
	  gcmp->aname[base[i]] = NULL;
	  for(j=0; j<3; ++j) (g->x + 3*i)[j] = (gcmp->x + 3*base[i])[j];
	  if(gcmp->flags != NULL)
	    g->flags[i] = gcmp->flags[base[i]];
	}
	Gfree(gcmp);
	myfree(base);
  }

  if(flag >= 0 || Flag >= 0)
  {
  	if(g->flags == NULL)
  	{
  		g->flags = imalloc(g->n);
  		for(i=0; i<(g->n); ++i) g->flags[i] = 0;
	}
  	for(i=0; i<(g->n); ++i)
	{
	  if(Flag >= 0) g->flags[i] &= Flag;
	  if(flag >= 0) g->flags[i] |= flag;
	}
  }

  if(Orient) Gstdorient(g);

  if(strcmp(ofn, "-")) {
    if((f = fopen(ofn, "w"))==NULL) {
      fprintf(stderr, "%s: cannot open %s for write\n", argv[0], ofn);
      exit(1);
    }
  } else {
    f = stdout;
  }
  switch(oformat) {
    case BOX: Gprintbox(f, g); break;
    case XYZ: GprintXYZ(f, g); break;
    case GENERAL:
    default:
      Gprint(f, g); break;
  }
  if(f != stdout) fclose(f);

  if(Mflag) {
    double mom[3], xcam[3], axes[3*3];
    Gmoment(g, NULL, mom, xcam, axes);
    printf("I = {%g,%g,%g}\n", mom[0], mom[1], mom[2]);
  }

  if(nflag > 0.0) {
    NeighborList *nl = nlalloc(nflag);
    newnl(g, nl);
    fprintnl(stdout, g, nl, g->n);
    nlfree(nl);
  }

  exit(0);
}
